#include "nbody.h"

#define	REAL3 MACROCONCAT(REAL, 3, )
#define	REAL4 MACROCONCAT(REAL, 4, )

int interactionBodyBody(REAL3 *ai, REAL4 bi, REAL4 bj, REAL eps2, REAL kappa) 
{
    REAL3 r;
    
    r.x = bi.x - bj.x;
    r.y = bi.y - bj.y;
    r.z = bi.z - bj.z;

    REAL r2 = r.x * r.x + r.y * r.y + r.z * r.z + eps2;	// r^2 
    REAL rinv = rsqrt(r2);								// sqrt^{-1}
	REAL a_rinv = kappa * bj.w * rinv * rinv * rinv;    // a r^{-1}

    (*ai).x += a_rinv * r.x;
    (*ai).y += a_rinv * r.y;
    (*ai).z += a_rinv * r.z;

    return 0;
}

__kernel void interactionBodyAll(__global REAL3 *a, __global REAL4 *b, int bodies, REAL eps2, REAL kappa,
                                 __local REAL4 *blocal)
{		
	int i = get_global_id(0);
    int ilocal = get_local_id(0);
    int tilesize = get_local_size(0);
    int tiles = (bodies - 1) / tilesize + 1;
    int bodiesleft = bodies;
	
    REAL3 ai = (REAL3)(0.0, 0.0, 0.0);

    for (int t = 0; t < tiles; t++)
    {
        if (ilocal < bodiesleft)
            blocal[ilocal] = b[t * tilesize + ilocal];
        barrier(CLK_LOCAL_MEM_FENCE);
        for (int j = 0; j < min(tilesize, bodiesleft); j++)
            interactionBodyBody(&ai, b[i], blocal[j], eps2, kappa);
        barrier(CLK_LOCAL_MEM_FENCE);
        bodiesleft -= tilesize;
    }

    if (i < bodies)
        a[i] = ai;
}

__kernel void updateBody(__global REAL4 *b, __global REAL3 *v, __global REAL3 *a, int bodies, REAL dt)			
{		
	int i = get_global_id(0);
	
	if (i < bodies)
	{
		b[i].x += v[i].x * dt + 0.5 * a[i].x * dt * dt;
		b[i].y += v[i].y * dt + 0.5 * a[i].y * dt * dt;
		b[i].z += v[i].z * dt + 0.5 * a[i].z * dt * dt;
		
		v[i] += a[i] * dt;
	}
}
